/*
 * dpmutils/dpm405lp.c  Simple runtime DPM policy creation for IBM 405LP
 *  er the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Copyright (C) 2003, International Business Machines Corporation
 * All Rights Reserved.
 *
 * Bishop Brock
 * IBM Research, Austin Center for Low-Power Computing
 * bcbrock@us.ibm.com
 * April, 2003
 *
 * Inspired by a utility written by Bob Paulsen
 */

/*
  This simple utility uses the DPM system call interface to create operating
  points, classes and policies for the IBM405LsP reference designs "Beech" and
  "Arctic2".  The utility can also optionally activate a policy.

  dpm405lp can be used for testing, and also for cases where a policy manager
  is not required in the system.  dpm405lp creates a class and policy for every
  operating point. Simple idle-scaling policies are also defined. If a
  load-based policy manager is required then the 'spmt' utility should be used.

  Suggested use:

  $ dpm405lp -a -p idle-vfscale
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include "linux/config.h"
#include "linux/dpm.h"
#include "linux/unistd.h"

#include "my_threads.h"




/*
 * Key to operating point parameters.  For a more detailed description refer to
 * the Linux kernel file arch/ppc/platforms/ibm405lp_dpm.h
 *
 * A=voltage (-1 = default)
 * B=csel
 * C=pll mult
 * D=pll div
 * E=cpu/plb
 * F=cpu/sysclkout
 * G=cpu/uart
 * H=plb/opb
 * I=plb/ebc
 * J=plb/pxl
 * K=scsel
 * L=pcsel
 * M=sd_refresh (-1 = default)
 * N=sd_timing (-1 = default) */

struct op {
  char *name;
  dpm_md_pp_t pp[DPM_PP_NBR];
};

/* Operating points for Beech using the VGA panel.  Since Beech has a separate
   serial clock, it has a wider range of operating points than Arctic2.  The
   Beech pixel clock is programmed from 16.66 to 25 MHz. EBC/OPB are normally
   defined between 25 and 33 MHz. */
struct op beech_points[] = {
        /*                     A  B   C   D  E   F   G  H  I   J  K  L   M   N */

        /* PLL bypass points */
        { "relock",        {  -1, 0, 16, 16, 1, 32, 32, 1, 1,  2, 0, 0, -1, -1} },

        /* VCO = 533 MHz points */
        { "533/33/33",     {  -1, 1, 16, 16, 1, 32, 32, 1, 1,  2, 0, 0, -1, -1} },
        { "533/33/33@maxv",{1800, 1, 16, 16, 1, 32, 32, 1, 1,  2, 0, 0, -1, -1} },

        { "66/33",         {  -1, 1, 16,  8, 2, 32, 32, 1, 1,  2, 0, 0, -1, -1} },
        { "66/33@maxv",    {1800, 1, 16,  8, 2, 32, 32, 1, 1,  2, 0, 0, -1, -1} },

		/* FREQ_POINTS[1] */
        { "66/66",         {1100, 1, 16,  8, 1, 32, 32, 2, 2,  3, 0, 0, -1, -1} },
        { "66/66/16",      {1100, 1, 16,  8, 1, 32, 32, 2, 2,  3, 0, 0, -1, -1} },
        { "66/33/16",      {1100, 1, 16,  8, 2, 32, 32, 2, 2,  3, 0, 0, -1, -1} },  /* Santhosh: New OP */
        { "33/33/16",      {1000, 1, 16, 16, 1, 32, 32, 1, 1,  2, 0, 0, -1, -1} },
	{ "44/44/16",      {1000, 1, 16, 12, 1, 32, 32, 1, 1,  2, 0, 0, -1, -1} },
        { "133/33",        {  -1, 1, 16,  4, 4, 32, 32, 1, 1,  2, 0, 0, -1, -1} },
        { "133/66",        {  -1, 1, 16,  4, 2, 32, 32, 2, 2,  3, 0, 0, -1, -1} },


		/* FREQ_POINTS[3] */
        { "133/133",       {1300, 1, 16,  4, 1, 32, 32, 4, 4,  6, 0, 0, -1, -1} },
        { "133/133/16",    {1300, 1, 16,  4, 1, 32, 32, 4, 4,  6, 0, 0, -1, -1} },
        { "133/66/16",     {1300, 1, 16,  4, 2, 32, 32, 4, 4,  6, 0, 0, -1, -1} },  /* Santhosh: New OP */
        { "133/33/16",     {1300, 1, 16,  4, 4, 32, 32, 4, 4,  6, 0, 0, -1, -1} },  /* Santhosh: New OP */

        { "266/33",        {  -1, 1, 16,  2, 8, 32, 32, 1, 1,  2, 0, 0, -1, -1} },
        { "266/66",        {  -1, 1, 16,  2, 4, 32, 32, 2, 2,  3, 0, 0, -1, -1} },


		/* FREQ_POINTS[7] */
        { "266/133",       {1700, 1, 16,  2, 2, 32, 32, 4, 4,  6, 0, 0, -1, -1} },
        { "266/133/16",    {1700, 1, 16,  2, 2, 32, 32, 4, 4,  6, 0, 0, -1, -1} },
        { "266/66/16",     {1700, 1, 16,  2, 4, 32, 32, 4, 4,  6, 0, 0, -1, -1} },  /* Santhosh: New OP */
        { "266/33/16",     {1700, 1, 16,  2, 8, 32, 32, 4, 4,  6, 0, 0, -1, -1} },  /* Santhosh: New OP */

        /* VCO = 667 MHz points */
        { "83/83",         {  -1, 1, 20,  8, 1, 32, 32, 3, 3,  4, 0, 0, -1, -1} },

        { "166/33",        {  -1, 1, 20,  4, 5, 32, 32, 1, 1,  2, 0, 0, -1, -1} },
        { "166/56",        {  -1, 1, 20,  4, 4, 32, 32, 2, 2,  3, 0, 0, -1, -1} },


		/* FREQ_POINTS[4] */
        { "166/83",        {1400, 1, 20,  4, 2, 32, 32, 3, 3,  4, 0, 0, -1, -1} },

        /* VCO = 800 MHz */
        { "800/33/33",     {  -1, 1, 24, 24, 1, 32, 32, 1, 1,  2, 0, 0, -1, -1} },
        { "800/33/33@maxv",{1800, 1, 24, 24, 1, 32, 32, 1, 1,  2, 0, 0, -1, -1} },

        { "50/25",         {  -1, 1, 24, 16, 2, 32, 32, 1, 1,  1, 0, 0, -1, -1} },


		/* FREQ_POINTS[0] */
        { "50/50",         {1000, 1, 24, 16, 1, 32, 32, 2, 2,  2, 0, 0, -1, -1} },

        { "100/25",        {  -1, 1, 24,  8, 4, 32, 32, 1, 1,  1, 0, 0, -1, -1} },
        { "100/50",        {  -1, 1, 24,  8, 2, 32, 32, 2, 2,  2, 0, 0, -1, -1} },

		/* FREQ_POINTS[2] */
        { "100/100",       {1200, 1, 24,  8, 1, 32, 32, 4, 4,  4, 0, 0, -1, -1} },

        { "200/25",        {  -1, 1, 24,  4, 8, 32, 32, 1, 1,  1, 0, 0, -1, -1} },
        { "200/50",        {  -1, 1, 24,  4, 4, 32, 32, 2, 2,  2, 0, 0, -1, -1} },

		/* FREQ_POINTS[5] */
        { "200/100",       {1500, 1, 24,  4, 2, 32, 32, 4, 4,  4, 0, 0, -1, -1} },
	{ "200/100/24",    {1500, 1, 24,  4, 2, 32, 32, 4, 4,  4, 0, 0, -1, -1} },
        { "66/66/24",      {1100, 1, 24, 12, 1, 32, 32, 2, 2,  3, 0, 0, -1, -1} },


	/* VCO = 933 MHz */
        { "117/117",       {  -1, 1, 28,  8, 1, 32, 32, 4, 4,  5, 0, 0, -1, -1} },
        { "117/58",        {  -1, 1, 28,  8, 2, 32, 32, 2, 2,  3, 0, 0, -1, -1} },

		/* FREQ_POINTS[6] */
        { "233/117",       {1600, 1, 28,  4, 2, 32, 32, 4, 4,  5, 0, 0, -1, -1} },
        { "233/58",        {  -1, 1, 28,  4, 4, 32, 32, 2, 2,  3, 0, 0, -1, -1} },
};


extern char *optarg;
extern int optind, opterr, optopt;


#define check_syscall(name, call, abend)\
do { if ((call)) { perror(name); if (abend) abort(); } } while(0)

void
definitions(struct op *points, int npoints);

void
my_dpm_init(void )
{
	struct dpm_param params;
	int beech = 0, arctic2 = 0, help = 0, invalid = 0, define_policies = 0;
	char *policy_name = NULL;
	int c;
	struct op *points;
	int npoints;

	define_policies = 1;
	points = beech_points;
	npoints = sizeof(beech_points) / sizeof(struct op);
	/* Terminate DPM */
	check_syscall("DPM_TERMINATE", sys_dpm(DPM_TERMINATE, NULL), 1);

	/* Activate DPM */
	check_syscall("DPM_INIT", sys_dpm(DPM_INIT, NULL), 1);

	/* Define everything is required */
	if (define_policies)
		definitions(points, npoints);


	return;

}

#ifdef SYNC
void
my_set_freq(char* policy_name)
{

	struct dpm_param params;

	/*-- for debugging purpose --*/
	mytrace("\n***********\nset freq. to %s\n********\n", policy_name);

	params.name = policy_name;
	check_syscall("DPM_SET_POLICY", sys_dpm(DPM_SET_POLICY, &params), 0);
	
    return;
}


#else

void
my_set_policy(char* policy_name)
{

	struct dpm_param params;

	/*-- for debugging purpose --*/
	mytrace("\n***********\nset policy to %s\n********\n", policy_name);

	params.name = policy_name;
	check_syscall("DPM_SET_POLICY", sys_dpm(DPM_SET_POLICY, &params), 0);
	
    return;
}



void
my_set_freq(dpm_state_t task_state)
{

       
	struct dpm_param params;

	if(task_state==100){
		params.name = "200/100/24";
		check_syscall("DPM_SET_POLICY", sys_dpm(DPM_SET_POLICY, &params), 0);
	}else{

		/*-- for debugging purpose --*/
		mytrace("\n***********\nset freq. to %d\n********\n", task_state);

		/* if the frequency level does not follow the order of 33,44,66,133,266,
		   need to change the task_state mapping, too*/
		params.task_state = task_state;

		params.pid = 0;
		check_syscall("DPM_SET_TASK_STATE",
			      sys_dpm(DPM_SET_TASK_STATE, &params), 0);
	}

    return;
}
#endif


void
definitions(struct op *points, int npoints)
{
	int i, j;
	struct dpm_param params;
	char *class[1];
	char *policy[DPM_STATES];

	/* Create operating points */

	for (i = 0; i < npoints; i++) {
		params.name = points[i].name;
		params.pp = points[i].pp;
		check_syscall("DPM_CREATE_OPT",
			      sys_dpm(DPM_CREATE_OPT, &params), 0);
	}

	/* Create classes */

	for (i = 0; i < npoints; i++) {
		params.name = points[i].name;
		class[0] = points[i].name;
		params.m_names = class;
		params.m_count = 1;
		check_syscall("DPM_CREATE_CLASS",
			      sys_dpm(DPM_CREATE_CLASS, &params), 0);
	}

	/* Create uniform policies */

	for (i = 0; i < npoints; i++) {
		params.name = points[i].name;
		params.m_names = policy;
		for (j = 0; j < DPM_STATES; j++) {
			if (j == DPM_RELOCK_STATE)
				policy[j] = "relock";
			else
				policy[j] = points[i].name;
		}
		check_syscall("DPM_CREATE_POLICY",
			      sys_dpm(DPM_CREATE_POLICY, &params), 0);
	}

	/* Create special policies */

	/* "idle-vfscale" Voltage and frequency scale at idle
	   Idle = 533/33/33, Idle-Task = 66/33, Task = 266/133 */

	params.name = "idle-vfscale";
	params.m_names = policy;
	for (j = 0; j < DPM_STATES; j++) {
		if (j == DPM_RELOCK_STATE)
			policy[j] = "relock";
		else if (j == DPM_IDLE_STATE)
			policy[j] = "533/33/33";
		else if (j == DPM_IDLE_TASK_STATE)
			policy[j] = "66/33";
		else
			policy[j] = "266/133";
	}
	check_syscall("DPM_CREATE_POLICY",
		      sys_dpm(DPM_CREATE_POLICY, &params), 0);

	/* "idle-fscale" Frequency scale only at idle
	   Idle = 533/33/33@maxv, Idle-Task = Task = 266/133 */

	params.name = "idle-fscale";
	params.m_names = policy;
	for (j = 0; j < DPM_STATES; j++) {
		if (j == DPM_RELOCK_STATE)
			policy[j] = "relock";
		else if (j == DPM_IDLE_STATE)
			policy[j] = "533/33/33@maxv";
		else
			policy[j] = "266/133";
	}
	check_syscall("DPM_CREATE_POLICY",
		      sys_dpm(DPM_CREATE_POLICY, &params), 0);

	/* "idle-fscale-indirect" Frequency scale only at idle, but use the
	   idle-task state

	   Idle = 533/33/33@maxv, Idle-Task = 66/63@maxv, Task = 266/133 */

	params.name = "idle-fscale-indirect";
	params.m_names = policy;
	for (j = 0; j < DPM_STATES; j++) {
		if (j == DPM_RELOCK_STATE)
			policy[j] = "relock";
		else if (j == DPM_IDLE_STATE)
			policy[j] = "533/33/33@maxv";
		else if (j == DPM_IDLE_TASK_STATE)
			policy[j] = "66/33@maxv";
		else
			policy[j] = "266/133";
	}
	check_syscall("DPM_CREATE_POLICY",
		      sys_dpm(DPM_CREATE_POLICY, &params), 0);

#ifndef SYNC
	/* "DVS-533/16 for task states 0..3, o/w idle-vfscale */

	params.name = "DVS-533/16";
	params.m_names = policy;
	for (j = 0; j < DPM_STATES; j++) {
		if (j == DPM_RELOCK_STATE)
			policy[j] = "relock";
		else if (j == DPM_IDLE_STATE)
			policy[j] = "533/33/33";
		else if (j == DPM_IDLE_TASK_STATE)
			policy[j] = "66/33";
		else if (j == DPM_TASK_STATE) // state 0
			policy[j] = "33/33/16";
		else if (j == DPM_TASK_STATE + 1) // state 1
			policy[j] = "44/44/16";
		else if (j == DPM_TASK_STATE + 2) // state 1
			policy[j] = "66/66/16";
		else if (j == DPM_TASK_STATE + 3) // state 2
			policy[j] = "133/133/16";
		else if (j == DPM_TASK_STATE + 4) // state 3
			policy[j] = "266/133/16";
		else
			policy[j] = "266/133/16";
		
	}
	check_syscall("DPM_CREATE_POLICY",
		      sys_dpm(DPM_CREATE_POLICY, &params), 0);
#endif
}

/*
 * Local variables:
 * c-basic-offset: 8
 * End:
 */



